@@ -22,9 +22,10 @@
 //      LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 //      OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 //      THE SOFTWARE.
+use anyhow::anyhow;
 use serde;
 use serde::{Deserialize, Serialize};
-use serde_json::Error;
+use serde_json::error::Category;
 
 #[derive(Deserialize, Debug, PartialEq)]
 #[serde(tag = "type")]
@@ -35,6 +36,7 @@ pub(crate) enum Test {
         test_count: Option<usize>,
         passed: Option<usize>,
         failed: Option<usize>,
+        errors: Option<usize>,
         allowed_fail: Option<usize>,
         ignored: Option<usize>,
         measured: Option<usize>,
@@ -59,6 +61,8 @@ pub(crate) struct ParsedTestSuite {
     pub(crate) passed: usize,
     #[serde(rename = "failures")]
     pub(crate) failed: usize,
+    #[serde(rename = "errors")]
+    pub(crate) errors: usize,
     pub(crate) allowed_fail: usize,
     #[serde(rename = "skipped")]
     pub(crate) ignored: usize,
@@ -86,13 +90,23 @@ pub(crate) enum TestStatus {
     Failed,
 }
 
+pub(crate) trait TestParser {
+    type Error: Into<anyhow::Error>;
+
+    fn multi_line(&self) -> bool;
+    fn parse(&mut self, text: &str) -> Result<Option<ParsedTestSuite>, Self::Error>;
+    fn reset(&mut self);
+}
+
 impl Test {
+    #[cfg(test)]
     pub(crate) fn new_suite_started(event: String, test_count: usize) -> Test {
         Test::Suite {
             event,
             test_count: Some(test_count),
             passed: None,
             failed: None,
+            errors: None,
             allowed_fail: None,
             ignored: None,
             measured: None,
@@ -101,10 +115,12 @@ impl Test {
         }
     }
 
+    #[cfg(test)]
     pub(crate) fn new_suite_ok(
         event: String,
         passed: usize,
         failed: usize,
+        errors: Option<usize>,
         allowed_fail: usize,
         ignored: usize,
         measured: usize,
@@ -116,6 +132,7 @@ impl Test {
             test_count: None,
             passed: Some(passed),
             failed: Some(failed),
+            errors,
             allowed_fail: Some(allowed_fail),
             ignored: Some(ignored),
             measured: Some(measured),
@@ -125,29 +142,51 @@ impl Test {
     }
 }
 
-pub(crate) struct TestParser {
+pub(crate) struct CargoTestParserOptions {
+    pub(crate) ignore_parse_errors: bool,
+}
+
+impl Default for CargoTestParserOptions {
+    fn default() -> Self {
+        Self {
+            ignore_parse_errors: false,
+        }
+    }
+}
+
+pub(crate) struct CargoTestParser {
+    opts: CargoTestParserOptions,
     header: Option<Test>,
     bottom: Option<Test>,
     tests: Vec<ParsedTest>,
 }
 
-impl TestParser {
-    pub(crate) fn new() -> TestParser {
-        TestParser {
+impl CargoTestParser {
+    pub(crate) fn new(opts: CargoTestParserOptions) -> CargoTestParser {
+        CargoTestParser {
+            opts,
             header: None,
             bottom: None,
             tests: vec![],
         }
     }
+}
 
-    pub(crate) fn reset(&mut self) {
-        self.header = None;
-        self.bottom = None;
-        self.tests = vec![];
+impl TestParser for CargoTestParser {
+    type Error = anyhow::Error;
+
+    fn multi_line(&self) -> bool {
+        true
     }
 
-    pub(crate) fn parse(&mut self, text: &String) -> Result<Option<ParsedTestSuite>, Error> {
-        let test = parse_test(text.as_str())?;
+    fn parse(&mut self, text: &str) -> Result<Option<ParsedTestSuite>, Self::Error> {
+        let test = parse_test(text, self.opts.ignore_parse_errors)?;
+        let test = if let Some(t) = test {
+            t
+        } else {
+            return Ok(None);
+        };
+
         match test {
             Test::Suite { ref event, .. } => {
                 if event.to_string() == "started".to_string() {
@@ -196,6 +235,7 @@ impl TestParser {
             test_count: _,
             ref passed,
             ref failed,
+            ref errors,
             ref allowed_fail,
             ref ignored,
             ref measured,
@@ -225,7 +265,8 @@ impl TestParser {
                     test_count: test_count.unwrap(),
                     passed: passed.unwrap(),
                     failed: failed.unwrap(),
-                    allowed_fail: allowed_fail.unwrap(),
+                    errors: errors.unwrap_or(0),
+                    allowed_fail: allowed_fail.unwrap_or(0),
                     ignored: ignored.unwrap(),
                     measured: measured.unwrap(),
                     filtered_out: filtered_out.unwrap(),
@@ -237,8 +278,28 @@ impl TestParser {
 
         Ok(None)
     }
+
+    fn reset(&mut self) {
+        self.header = None;
+        self.bottom = None;
+        self.tests = vec![];
+    }
 }
 
-pub(crate) fn parse_test(str: &str) -> Result<Test, Error> {
-    serde_json::from_str(str)
+pub(crate) fn parse_test(
+    str: &str,
+    ignore_parse_error: bool,
+) -> Result<Option<Test>, anyhow::Error> {
+    if str.starts_with("{") {
+        let parsed = serde_json::from_str::<'_, Test>(str);
+        match parsed {
+            Ok(p) => Ok(Some(p)),
+            Err(e) => match e.classify() {
+                Category::Data if ignore_parse_error => return Ok(None),
+                _ => Err(anyhow!(e).context(format!("failed to parse test text: {}", str))),
+            },
+        }
+    } else {
+        Ok(None)
+    }
 }
